@@ -1,3 +1,4 @@
+# frozen_string_literal: true
 ##
 # BasicSpecification is an abstract class which implements some common code
 # used by both Specification and StubSpecification.
@@ -22,15 +23,34 @@ class Gem::BasicSpecification
   ##
   # The path this gemspec was loaded from.  This attribute is not persisted.
 
-  attr_reader :loaded_from
+  attr_accessor :loaded_from
 
   ##
   # Allows correct activation of git: and path: gems.
 
   attr_writer :full_gem_path # :nodoc:
 
+  def initialize
+    internal_init
+  end
+
   def self.default_specifications_dir
-    File.join(Gem.default_dir, "specifications", "default")
+    Gem.default_specifications_dir
+  end
+
+  class << self
+
+    extend Gem::Deprecate
+    deprecate :default_specifications_dir, "Gem.default_specifications_dir", 2020, 02
+
+  end
+
+  ##
+  # The path to the gem.build_complete file within the extension install
+  # directory.
+
+  def gem_build_complete_path # :nodoc:
+    File.join extension_dir, 'gem.build_complete'
   end
 
   ##
@@ -46,54 +66,46 @@ class Gem::BasicSpecification
   # eg: /usr/local/lib/ruby/gems/1.8
 
   def base_dir
-    return Gem.dir unless loaded_from
-    @base_dir ||= if default_gem? then
-                    File.dirname File.dirname File.dirname loaded_from
-                  else
-                    File.dirname File.dirname loaded_from
-                  end
+    raise NotImplementedError
   end
 
   ##
   # Return true if this spec can require +file+.
 
-  def contains_requirable_file? file
-    if instance_variable_defined?(:@ignored) or
-       instance_variable_defined?('@ignored') then
+  def contains_requirable_file?(file)
+    if @ignored
       return false
-    elsif missing_extensions? then
+    elsif missing_extensions?
       @ignored = true
 
-      warn "Ignoring #{full_name} because its extensions are not built.  " +
-           "Try: gem pristine #{full_name}"
+      if RUBY_ENGINE == platform || Gem::Platform.local === platform
+        warn "Ignoring #{full_name} because its extensions are not built. " +
+          "Try: gem pristine #{name} --version #{version}"
+      end
+
       return false
     end
 
-    suffixes = Gem.suffixes
-
-    full_require_paths.any? do |dir|
-      base = "#{dir}/#{file}"
-      suffixes.any? { |suf| File.file? "#{base}#{suf}" }
-    end
+    have_file? file, Gem.suffixes
   end
 
   def default_gem?
     loaded_from &&
-      File.dirname(loaded_from) == self.class.default_specifications_dir
+      File.dirname(loaded_from) == Gem.default_specifications_dir
   end
 
   ##
   # Returns full path to the directory where gem's extensions are installed.
 
   def extension_dir
-    @extension_dir ||= File.expand_path File.join(extensions_dir, full_name)
+    @extension_dir ||= File.expand_path(File.join(extensions_dir, full_name)).untaint
   end
 
   ##
   # Returns path to the extensions directory.
 
   def extensions_dir
-    @extensions_dir ||= Gem.default_ext_dir_for(base_dir) ||
+    Gem.default_ext_dir_for(base_dir) ||
       File.join(base_dir, 'extensions', Gem::Platform.local.to_s,
                 Gem.extension_api_version)
   end
@@ -102,7 +114,7 @@ class Gem::BasicSpecification
     # TODO: also, shouldn't it default to full_name if it hasn't been written?
     path = File.expand_path File.join(gems_dir, full_name)
     path.untaint
-    path if File.directory? path
+    path
   end
 
   private :find_full_gem_path
@@ -122,10 +134,10 @@ class Gem::BasicSpecification
   # default Ruby platform.
 
   def full_name
-    if platform == Gem::Platform::RUBY or platform.nil? then
-      "#{name}-#{version}".untaint
+    if platform == Gem::Platform::RUBY or platform.nil?
+      "#{name}-#{version}".dup.untaint
     else
-      "#{name}-#{version}-#{platform}".untaint
+      "#{name}-#{version}-#{platform}".dup.untaint
     end
   end
 
@@ -134,13 +146,46 @@ class Gem::BasicSpecification
   # activated.
 
   def full_require_paths
-    full_paths = raw_require_paths.map do |path|
-      File.join full_gem_path, path
+    @full_require_paths ||=
+    begin
+      full_paths = raw_require_paths.map do |path|
+        File.join full_gem_path, path.untaint
+      end
+
+      full_paths << extension_dir if have_extensions?
+
+      full_paths
     end
+  end
 
-    full_paths.unshift extension_dir unless @extensions.nil? || @extensions.empty?
+  ##
+  # The path to the data directory for this gem.
 
-    full_paths
+  def datadir
+    # TODO: drop the extra ", gem_name" which is uselessly redundant
+    File.expand_path(File.join(gems_dir, full_name, "data", name)).untaint
+  end
+
+  ##
+  # Full path of the target library file.
+  # If the file is not in this gem, return nil.
+
+  def to_fullpath(path)
+    if activated?
+      @paths_map ||= {}
+      @paths_map[path] ||=
+      begin
+        fullpath = nil
+        suffixes = Gem.suffixes
+        suffixes.find do |suf|
+          full_require_paths.find do |dir|
+            File.file?(fullpath = "#{dir}/#{path}#{suf}")
+          end
+        end ? fullpath : nil
+      end
+    else
+      nil
+    end
   end
 
   ##
@@ -156,23 +201,14 @@ class Gem::BasicSpecification
   # gem directory. eg: /usr/local/lib/ruby/1.8/gems
 
   def gems_dir
-    # TODO: this logic seems terribly broken, but tests fail if just base_dir
-    @gems_dir ||= File.join(loaded_from && base_dir || Gem.dir, "gems")
+    raise NotImplementedError
   end
 
-  ##
-  # Set the path the Specification was loaded from. +path+ is converted to a
-  # String.
-
-  def loaded_from= path
-    @loaded_from   = path && path.to_s
-
+  def internal_init # :nodoc:
     @extension_dir = nil
-    @extensions_dir = nil
-    @full_gem_path         = nil
-    @gem_dir               = nil
-    @gems_dir              = nil
-    @base_dir              = nil
+    @full_gem_path = nil
+    @gem_dir = nil
+    @ignored = nil
   end
 
   ##
@@ -190,7 +226,7 @@ class Gem::BasicSpecification
   end
 
   def raw_require_paths # :nodoc:
-    Array(@require_paths)
+    raise NotImplementedError
   end
 
   ##
@@ -211,7 +247,7 @@ class Gem::BasicSpecification
   #   spec.require_path = '.'
 
   def require_paths
-    return raw_require_paths if @extensions.nil? || @extensions.empty?
+    return raw_require_paths unless have_extensions?
 
     [extension_dir].concat raw_require_paths
   end
@@ -223,8 +259,8 @@ class Gem::BasicSpecification
   def source_paths
     paths = raw_require_paths.dup
 
-    if @extensions then
-      ext_dirs = @extensions.map do |extension|
+    if have_extensions?
+      ext_dirs = extensions.map do |extension|
         extension.split(File::SEPARATOR, 2).first
       end.uniq
 
@@ -234,6 +270,34 @@ class Gem::BasicSpecification
     paths.uniq
   end
 
+  ##
+  # Return all files in this gem that match for +glob+.
+
+  def matches_for_glob(glob) # TODO: rename?
+    # TODO: do we need these?? Kill it
+    glob = File.join(self.lib_dirs_glob, glob)
+
+    Dir[glob].map { |f| f.untaint } # FIX our tests are broken, run w/ SAFE=1
+  end
+
+  ##
+  # Returns a string usable in Dir.glob to match all requirable paths
+  # for this spec.
+
+  def lib_dirs_glob
+    dirs = if self.raw_require_paths
+             if self.raw_require_paths.size > 1
+               "{#{self.raw_require_paths.join(',')}}"
+             else
+               self.raw_require_paths.first
+             end
+           else
+             "lib" # default value for require_paths for bundler/inline
+           end
+
+    "#{self.full_gem_path}/#{dirs}".dup.untaint
+  end
+
   ##
   # Return a Gem::Specification from this gem
 
@@ -256,5 +320,24 @@ class Gem::BasicSpecification
     raise NotImplementedError
   end
 
-end
+  def this; self; end
 
+  private
+
+  def have_extensions?; !extensions.empty?; end
+
+  def have_file?(file, suffixes)
+    return true if raw_require_paths.any? do |path|
+      base = File.join(gems_dir, full_name, path.untaint, file).untaint
+      suffixes.any? { |suf| File.file? base + suf }
+    end
+
+    if have_extensions?
+      base = File.join extension_dir, file
+      suffixes.any? { |suf| File.file? base + suf }
+    else
+      false
+    end
+  end
+
+end
