@@ -216,6 +216,12 @@ class TestRegexp < Test::Unit::TestCase
     assert_not_include(local_variables, :nil, "[ruby-dev:32675]")
   end
 
+  def test_assign_named_capture_to_const
+    %W[C \u{1d402}].each do |name|
+      assert_equal(:ok, Class.new.class_eval("#{name} = :ok; /(?<#{name}>.*)/ =~ 'ng'; #{name}"))
+    end
+  end
+
   def test_assign_named_capture_trace
     bug = '[ruby-core:79940] [Bug #13287]'
     assert_normal_exit("#{<<-"begin;"}\n#{<<-"end;"}", bug)
@@ -533,7 +539,7 @@ class TestRegexp < Test::Unit::TestCase
   end
 
   def test_match
-    assert_nil(//.match(nil))
+    assert_raise(TypeError) { //.match(nil) }
     assert_equal("abc", /.../.match(:abc)[0])
     assert_raise(TypeError) { /.../.match(Object.new)[0] }
     assert_equal("bc", /../.match('abc', 1)[0])
@@ -555,10 +561,10 @@ class TestRegexp < Test::Unit::TestCase
     /backref/ =~ 'backref'
     # must match here, but not in a separate method, e.g., assert_send,
     # to check if $~ is affected or not.
-    assert_equal(false, //.match?(nil))
     assert_equal(true, //.match?(""))
     assert_equal(true, /.../.match?(:abc))
     assert_raise(TypeError) { /.../.match?(Object.new) }
+    assert_raise(TypeError) { //.match?(nil) }
     assert_equal(true, /b/.match?('abc'))
     assert_equal(true, /b/.match?('abc', 1))
     assert_equal(true, /../.match?('abc', 1))
@@ -944,24 +950,26 @@ class TestRegexp < Test::Unit::TestCase
   end
 
   def test_cclass_R
-    assert_match /\A\R\z/, "\r"
-    assert_match /\A\R\z/, "\n"
-    assert_match /\A\R\z/, "\r\n"
+    assert_match(/\A\R\z/, "\r")
+    assert_match(/\A\R\z/, "\n")
+    assert_match(/\A\R\z/, "\r\n")
   end
 
   def test_cclass_X
-    assert_match /\A\X\z/, "\u{20 200d}"
-    assert_match /\A\X\z/, "\u{600 600}"
-    assert_match /\A\X\z/, "\u{600 20}"
-    assert_match /\A\X\z/, "\u{261d 1F3FB}"
-    assert_match /\A\X\z/, "\u{1f600}"
-    assert_match /\A\X\z/, "\u{20 308}"
-    assert_match /\A\X\X\z/, "\u{a 308}"
-    assert_match /\A\X\X\z/, "\u{d 308}"
-    assert_match /\A\X\z/, "\u{1F477 1F3FF 200D 2640 FE0F}"
-    assert_match /\A\X\z/, "\u{1F468 200D 1F393}"
-    assert_match /\A\X\z/, "\u{1F46F 200D 2642 FE0F}"
-    assert_match /\A\X\z/, "\u{1f469 200d 2764 fe0f 200d 1f469}"
+    assert_match(/\A\X\z/, "\u{20 200d}")
+    assert_match(/\A\X\z/, "\u{600 600}")
+    assert_match(/\A\X\z/, "\u{600 20}")
+    assert_match(/\A\X\z/, "\u{261d 1F3FB}")
+    assert_match(/\A\X\z/, "\u{1f600}")
+    assert_match(/\A\X\z/, "\u{20 324}")
+    assert_match(/\A\X\X\z/, "\u{a 324}")
+    assert_match(/\A\X\X\z/, "\u{d 324}")
+    assert_match(/\A\X\z/, "\u{1F477 1F3FF 200D 2640 FE0F}")
+    assert_match(/\A\X\z/, "\u{1F468 200D 1F393}")
+    assert_match(/\A\X\z/, "\u{1F46F 200D 2642 FE0F}")
+    assert_match(/\A\X\z/, "\u{1f469 200d 2764 fe0f 200d 1f469}")
+
+    assert_warning('') {/\X/ =~ "\u{a0}"}
   end
 
   def test_backward
@@ -1067,6 +1075,9 @@ class TestRegexp < Test::Unit::TestCase
     assert_no_match(/^\p{age=3.0}$/u, "\u2754")
     assert_no_match(/^\p{age=2.0}$/u, "\u2754")
     assert_no_match(/^\p{age=1.1}$/u, "\u2754")
+
+    assert_no_match(/^\p{age=12.0}$/u, "\u32FF")
+    assert_match(/^\p{age=12.1}$/u, "\u32FF")
   end
 
   MatchData_A = eval("class MatchData_\u{3042} < MatchData; self; end")
@@ -1113,6 +1124,8 @@ class TestRegexp < Test::Unit::TestCase
 
     bug8151 = '[ruby-core:53649]'
     assert_warning(/\A\z/, bug8151) { Regexp.new('(?:[\u{33}])').to_s }
+
+    assert_warning(%r[/.*/\Z]) { Regexp.new("[\n\n]") }
   end
 
   def test_property_warn
